/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: val.c,v 2.14 2009/12/07 11:18:30 acau Exp $
 */


#include "system.h"





/*****************\
*                 *
*  Type checking  *
*                 *
\*****************/

void type_check(e,type)
NODEPTR e;
unsigned type;
{
    assert2 (node_type(e) == type,
            "found expression of type %s when expecting %s",
            print_name (e), Keyword[type]);
}

/******************************************\
*                                         *
*  checking  whether e is static variable *
*                                         *
\******************************************/

void static_check(e)
NODEPTR e;
{
    assert1 (node_type(e) == STATIC,
            "variable %s is of type state should be static",
             node_nm_name(e));
}

/********************************\
*                                *
*  Test whether val is defined   *
*                                *
\********************************/
int val_defined(e)
NODEPTR e;
{
    switch (node_type(e)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case LAMBDA:
            return (T);
        case LOC:
            return (pointer_defined ((e)));
        case LIST:
            return (list_defined (e));
        default:
            error_break1 ("value cannot be of type %s", print_name(e));
            return (F);
    }
}


int val_defined2(e)
NODEPTR e;
{
    switch (node_type(e)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case LAMBDA:
            return (T);
        case LOC:
            return (pointer_defined ((e)));
        case LIST:
            return (list_defined (e));
        default:
            return (F);
    }
}

int pointer_defined(p)
NODEPTR p;
{
    return (mem_defined1 (node_loc_tonext(p) ?
                         next_loc (node_loc_ptr(p)) :
                         curr_loc (node_loc_ptr(p))));
}


/* Check whether list is defined. */
int list_defined(list)
NODEPTR list;
{
    register int     current_size;
    register unsigned i;
    register NODEPTR base_loc;
    NODEPTR size_loc;
    
    if (list_size_defined (list)) {
        base_loc = list_base (list);
            /*printf("list_defined: base_loc=%p\n",base_loc);*/
        size_loc = list_size (list);
        current_size = get_int (size_loc);
        for (i = 0; (int)i < current_size; i += 1) {
            if (!mem_defined1 (base_loc + offset2 (i,node_mem_history(size_loc))))
                return (F);
            if (!val_defined (get_mem (base_loc + offset2 (i,node_mem_history(size_loc)))))
                return (F);
        }
        return (T);
    }
    else{
        return (F);
    }
}

int max_size(list)
NODEPTR list;
{
    NODEPTR size_loc;
    if (mem_defined1 (size_loc = list_size (list)) && is_static_mem (size_loc))
        return (get_int (size_loc));
    else
        return (node_blk_size(node_lst_blk(list)) - 1);
}

/* Check whether size of list is defined. */
int list_size_defined(list)
NODEPTR list;
{
    int     current_size;
    register NODEPTR size_loc = list_size (list);
    
    if (mem_defined1 (size_loc)) {
        current_size = get_int (size_loc);
        assert2d ((current_size >= 0) && (current_size <= max_size (list)),
                "current size of list is out of range [0,%d]: %d",
                max_size (list), current_size);
        return (T);
    }
    else {
        return (F);
    }
}



/********************************************************\
*                                                        *
*  Copying, comparing, assigning and outputting values.  *
*                                                        *
\********************************************************/

/* convert a value to a printable form */
void output_val(e)
NODEPTR e;
{
    switch (node_type(e)) {
        case NIL:
	    output_str ("nil");
	    break;
        case BOOL:
            output_str (e == True ? "true" : "false");
            break;
        case INT:
            output_int (node_ival(e), 0);
            break;
        case STRING:
            output_char ('"');
            output_str (node_strval(e));
            output_char ('"');
            break;
        case FLOAT:
            output_char ('$');
            output_fl (node_flval(e));
            output_char ('$');
            break;    
        case LIST:
            output_list (e);
            break;
        case LOC:
            output_char ('@');
            output_int ((int)cell_no (node_loc_ptr(e)), 0);
            break;
        case LAMBDA:
            output_lambda (e);
            break;   
        default:
            error_break1 ("expression of type %s cannot be output",
                    print_name (e));
    }
}

void foutput_val(e)
NODEPTR e;
{
    switch (node_type(e)) {
        case NIL:
	    foutput_str ("nil");
	    break;
        case BOOL:
            foutput_str (e == True ? "true" : "false");
            break;
        case INT:
            foutput_int (node_ival(e), 0);
            break;
        case STRING:
            foutput_char ('"');
            foutput_str (node_strval(e));
            foutput_char ('"');
            break;
        case FLOAT:
            foutput_char ('$');
            foutput_fl (node_flval(e));
            foutput_char ('$');
            break;    
        case LIST:
            foutput_list (e);
            break;
        case LOC:
            foutput_char ('@');
            foutput_int ((int)cell_no (node_loc_ptr(e)), 0);
            break;
        case LAMBDA:
            foutput_lambda (e);
            break;   
        default:
            error_break1 ("expression of type %s cannot be output",
                    print_name (e));
    }
}


void output_list(list)
NODEPTR list;
{
    int     size;
    unsigned i = 0;
    NODEPTR loc,
            base_loc;

    base_loc = list_base (list);
        /*printf("output_list: base_loc=%p\n",base_loc);*/
    size = mem_defined1 (loc = list_size (list)) ? get_int (loc) : max_size (list);
    output_char ('[');
    while ((int)i < size) {
        if (mem_defined1 (loc = base_loc + offset2 (i, node_mem_history(loc))))
            output_val (get_mem (loc));
        else
            output_char ('?');
        if ((int)++i < size)
            output_char (',');
    }
    output_char (']');
}

void foutput_list(list)
NODEPTR list;
{
    int     size;
    unsigned i = 0;
    NODEPTR loc,
            base_loc;

    base_loc = list_base (list);
        /*printf("foutput_list: base_loc=%p\n",base_loc);*/
    size = mem_defined1 (loc = list_size (list)) ? get_int (loc) : max_size (list);
    foutput_char ('[');
    while ((int)i < size) {
        if (mem_defined1 (loc = base_loc + offset2 (i, node_mem_history(loc))))
            foutput_val (get_mem (loc));
        else
            foutput_char ('?');
        if ((int)++i < size)
            foutput_char (',');
    }
    foutput_char (']');
}


/* Return a copy of a value. Atoms need no copying. */
NODEPTR copy_val(val)
NODEPTR val;
{
    switch (node_type(val)) {
        case NIL:
        case BOOL:
        case INT:
        case LAMBDA:
        case STRING:
        case FLOAT:
        case LOC:
        case STRUCT:
            return (val);
        case LIST:
            return (copy_list(val));
        default:
            error_break1 ("cannot copy value of type %s", print_name(val));
            return (F);
    }
}

/* Return a copy of a value. Atoms need no copying. */
NODEPTR copy_val2(val)
NODEPTR val;
{
    switch (node_type(val)) {
        case NIL:
        case BOOL:
        case INT:
        case LAMBDA:
        case STRING:
        case FLOAT:
        case LOC:
        case STRUCT:
            return (val);
        case LIST:
            return (copy_list2(val));
        default:
            return (val);
    }
}

/* Make a copy of a list. */
/* The copy is defined in the CURRENT state. */
NODEPTR copy_list(list)
NODEPTR list;
{
    NODEPTR size_loc = list_size (list);
    NODEPTR base_loc = list_base (list);
    NODEPTR new_base;
    NODEPTR list1;
    int i, size;
    unsigned int history = node_mem_history(size_loc);

/*printf("***Copying List\n");*/
        /*printf("copy_list: base_loc=%p\n",base_loc);*/
    assert0 (list_defined (list),
            "attempt to copy an incompletely defined list");
    
    list1 = new_list1 (STATIC, STATIC, get_mem (size_loc), history);
/*    if (is_static_mem (base_loc) && is_static_mem (size_loc)) {*/
/*printf("***SAVED\n");*/
/*        return (list);*/
/*	}*/
    list = list1;
    new_base = curr_loc (list_base (list));
    size = get_int (size_loc);
    for (i = 0; i < size; i += 1)
        put_mem1 (new_base + offset2 (i,history),
                  copy_val (get_mem (base_loc + offset2 (i,history))));
    return (list);
}

/* Make a copy of a list. */
/* The copy is defined in the CURRENT state. */
NODEPTR copy_list2(list)
NODEPTR list;
{
    NODEPTR size_loc = list_size (list);
    NODEPTR base_loc = list_base (list);
    NODEPTR new_base;
    NODEPTR list1;
    int i, size;
    unsigned int history = node_mem_history(size_loc);
    

/*printf("***Copying List\n");*/
    assert0 (list_defined (list),
            "attempt to copy an incompletely defined list");
        /*printf("copy_list2: base_loc=%p\n",base_loc);*/
    list1 = new_list1 (STATIC, STATIC, get_mem (size_loc), history);
/*    if (is_static_mem (base_loc) && is_static_mem (size_loc)) {*/
/*printf("***SAVED\n");*/
/*        return (list);*/
/*	}*/
    list = list1;
    new_base = list_base (list);
    size = get_int (size_loc);
    for (i = 0; i < size; i += 1)
        put_mem1 (new_base + offset2 (i,history),
                 copy_val2 (get_mem (base_loc + offset2 (i,history))));
    return (list);
}

int string_size(s)
NODEPTR s;
{
    return (strlen (node_strval(s)));
}

NODEPTR substring(s,i,j)
NODEPTR s;
int i,j;
{
    return (string_node (substr (node_strval(s), i, j)));
}

NODEPTR append_string (s1, s2)
NODEPTR s1, s2;
{
    return (string_node (strapnd (node_strval(s1), node_strval(s2))));
}

/* Return the result of appending list2 to the end of list1. */
/* The new list is defined in the CURRENT state. */
NODEPTR append_list(list1,list2)
NODEPTR list1,list2;
{
    NODEPTR     new_base,
            base_loc1,
            base_loc2;
    int     size1,
            size2;
    unsigned i;
    unsigned int history;
    

    base_loc1 = list_base (list1);
        /*printf("append_list: base_loc1=%p\n",base_loc1);*/
    history = node_mem_history(base_loc1);
    size1 = get_int (list_size (list1));
    base_loc2 = list_base (list2);
        /*printf("append_list: base_loc2=%p\n",base_loc2);*/
    size2 = get_int (list_size (list2));
    list1 = new_list1 (STATIC, STATIC, int_node (size1 + size2), history);
    new_base = curr_loc (node_lst_base(list1));
        /*printf("append_list: new_base_loc2=%p\n",new_base);*/
    for (i = 0; (int)i < size1; i += 1)
        put_mem1 (new_base + offset2 (i, history),
                  copy_val (get_mem (base_loc1 + offset2 (i,history))));
    new_base += offset2 (i,history);
    for (i = 0; (int)i < size2; i += 1)
        put_mem1 (new_base + offset2 (i,history),
                  copy_val (get_mem (base_loc2 + offset2 (i,history))));
    return (list1);
}

/* Create a new list. */
/*
NODEPTR new_list(cell_type,size_type,n)
int cell_type,size_type;
NODEPTR n;
{
    NODEPTR blk = mem_block (cell_type, (unsigned)node_ival(n) + 1);
    NODEPTR size_loc = node_blk_base(blk),
        base_loc = size_loc + offset (1);

    new_loc (size_loc, size_type);
    if (size_type == STATIC)
        put_mem (size_loc, n);
    return (list_node (blk, size_loc, base_loc));
}
*/
/* Create a new list. */
NODEPTR new_list1(cell_type,size_type,n,history)
int cell_type,size_type;
NODEPTR n;
unsigned int history;
{
    NODEPTR blk = mem_block1 (cell_type, (unsigned)node_ival(n) + 1,history);
    NODEPTR size_loc = node_blk_base(blk),
        base_loc = size_loc + offset2 (1,history);
        /*printf("new_list1: base_loc=%p\n",base_loc);*/
    new_loc1 (size_loc, size_type, history);
    if (size_type == STATIC)
        put_mem1 (size_loc, n);
    return (list_node (blk, size_loc, base_loc));
}


/* Assign a value to a location */
/*
void assign_val (loc,val)
NODEPTR val;
register NODEPTR loc;
{
    switch (node_type(val)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case LAMBDA:
        case LOC:
            if (mem_defined (loc)) {
                if (RedundancyFlag || !equal_val (val, get_mem (loc))) {
		    if (loc == TempLoc)
		        error_break0 ("attempt to redefine constant");
		    else{
                        if (node_mem_isundefined(loc)){
                            printf("do we want this?\n");
                                put_mem (loc, val);
                        }
                        else{
                            error_break0 ("attempt to overwrite variable");
                        }
                    }
		} 
            }
            else
                put_mem (loc, val);
            break;
        case STRUCT:
            assign_struct (loc, node_lnk_1(val));
            break;
        case LIST:
            assign_list (loc, val);
            break;
        default:
            error_break1 ("cannot assign value of type %s", print_name (val));
    }
}
*/

/* Assign a value to a location */
void assign_val1 (loc,val)
NODEPTR val;
register NODEPTR loc;
{
        /*if (DebugLevel>2)
          printf("assign_val: history=%d\n",node_mem_history(loc));*/
    switch (node_type(val)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case LAMBDA:
        case LOC:
            if (mem_defined1 (loc)) {
                if (RedundancyFlag || !equal_val (val, get_mem (loc))) {
		    if (loc == TempLoc)
		        error_break0 ("attempt to redefine constant");
		    else{
                        if (node_mem_isundefined(loc)){
                            printf("do we want this?\n");
                                /*put_mem (loc, val);*/
                        }
                        else{
                            error_break0 ("attempt to overwrite variable");
                        }
                    }
		} 
            }
            else
                put_mem1 (loc, val);
            break;
        case STRUCT:
            assign_struct (loc, node_lnk_1(val));
            break;
        case LIST:
            assign_list (loc, val);
            break;
        default:
            error_break1 ("cannot assign value of type %s", print_name (val));
    }
}

void assign_struct(loc,struct1)
NODEPTR loc;
NODEPTR struct1;
{
    if (!mem_defined1 (loc))
        put_mem1 (loc, struct1);
    else {
	register int size2;
        register NODEPTR loc1;
        register NODEPTR loc2;
        NODEPTR struct2 = get_list (loc);
        type_check (struct2, LIST);
        struct1 = put_list (loc, struct1); 
        assert0 (max_size (struct1) >= max_size (struct2),
	         "attempt to redefine max. size of list");
        if (mem_defined1 (node_lst_size(struct2))) {
	    size2 = get_int (node_lst_size(struct2));
	    if (node_lst_size(struct1) != node_lst_size(struct2)) {
	        put_mem1 (node_lst_size(struct1), get_mem (node_lst_size(struct2)));
                make_indirect_mem (node_lst_size(struct2), node_lst_size(struct1));
	    }
            loc1 = node_lst_base(struct1);
            loc2 = node_lst_base(struct2);
	    if (loc1 != loc2) {
                while (size2-- > 0) {
	            if (mem_defined1 (loc2))
                        assign_val1 (loc1, get_mem (loc2));
                    make_indirect_mem (loc2, loc1);
                    loc1 += offset2(1, node_mem_history(loc1));
                    loc2 += offset2(1, node_mem_history(loc2));
		}
	    }
	}
    }
}

/* copy one list into another. */
void assign_list(loc,list)
NODEPTR loc;
NODEPTR list;
{
    int     current_size1,
            current_size2;
    register NODEPTR loc1;
    register NODEPTR loc2;
    NODEPTR size_loc1;
    NODEPTR size_loc2;
    NODEPTR list1;
    
    
    loc2 = list_base (list);
        /*printf("assign_list: base_list=%p\n",list_base (list));*/
    size_loc2 = list_size (list);
    current_size2 = get_int (size_loc2);
    if (mem_defined1 (loc)) {
        assert0 (is_list (list1 = get_list (loc)),
                "attempt to overwrite non-list with list");
        assert2d (max_size (list1) >= current_size2,
                "attempt to assign list of size %d to one of max size %d",
                current_size2, max_size (list1));
    }
    else {
        list1 = new_list1 (is_static_mem (loc) ? STATIC : STATE,
                           STATIC, get_mem (size_loc2), node_mem_history(loc));
        list1 = put_list (loc, list1);/* possibly adjusted. */
    }
    loc1 = list_base (list1);
        /*printf("assign_list: base_list1=%p\n",loc1);*/
    size_loc1 = list_size (list1);

    if (!mem_defined1 (size_loc1)) {
        current_size1 = current_size2;
        put_mem1 (size_loc1, int_node (current_size1));
    }
    else {
        assert2d ((current_size1 = get_int (size_loc1)) == current_size2,
                "attempt to assign list of size %d to one of size %d",
                current_size2, current_size1);
    }
    while (current_size2-- > 0) {
        if (mem_defined1(loc2))
            assign_val1 (loc1, get_mem (loc2));
	loc1 += offset2 (1,node_mem_history(loc1));
	loc2 += offset2 (1,node_mem_history(loc2));
    }
}

/* Test whether two values are equal. */
int equal_val(e1,e2)
NODEPTR e1,e2;
{
    if (node_type(e1) != node_type(e2))
        return (F);
    return (compare_val (e1, e2) == 0);
}


/* Compare two values. Return 0 if e1=e2, >0 if e1>e2, <0 if e1<e2. */
int compare_val(e1,e2)
NODEPTR e1,e2;
{
    unsigned type;

    if ((type = node_type(e1)) != node_type(e2))
        return (-1);
    switch (type) {
        case NIL:
	    return (0);
        case BOOL:
        case INT:
            return (node_ival(e1) - node_ival(e2));
        case LAMBDA:
            return (e1 - e2);
        case LOC:
                  return (compare_ptrs (e1, e2));
        case STRING:
            return (compare_strings (e1, e2));
        case FLOAT:
            return (node_flval(e1) - node_flval(e2));    
        case LIST:
            return (compare_lists (e1, e2));
        default:
            error_break1 ("cannot compare values of type %s", print_name (e1));
            return (F);
    }
}

int compare_strings(e1,e2)
NODEPTR e1,e2;
{
    return (strcmp (node_strval(e1), node_strval(e2)));
}

int compare_ptrs(e1,e2)
NODEPTR e1,e2;
{
    return (node_loc_tonext(e1) == node_loc_tonext(e2) ?
            node_loc_ptr(e1) - node_loc_ptr(e2) :
            node_loc_ptr(e1) - node_loc_ptr(e2));
}

int compare_lists(list1,list2)
NODEPTR list1,list2;
{
    int     size1,
            size2,
            answer;
    unsigned i;
    NODEPTR base_loc1,
            base_loc2;

    assert0 (list_defined (list1) && list_defined (list2),
            "attempt to compare an undefined list");
    base_loc1 = list_base (list1);
    base_loc2 = list_base (list2);
        /*printf("compare_lists: base_loc1=%p\n",base_loc1);*/
        /*printf("compare_lists: base_loc2=%p\n",base_loc2);*/
    size1 = get_int (list_size (list1));
    size2 = get_int (list_size (list2));
    if (size1 != size2)
        return (size1 - size2);
    for (i = 0; (int)i < size1; i += 1) {
        if ((answer = compare_val (get_mem (base_loc1 + offset2 (i,node_mem_history(base_loc1))),
                                            get_mem (base_loc2 + offset2 (i,node_mem_history(base_loc2))))) != 0)
            return (answer);
    }
    return (0);
}

